(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)

  
  
module type LABELED_GRAPH_BASE = sig
  type node_description
  type label

  val print_description : Format.formatter -> node_description -> unit
  val print_label : Format.formatter -> label -> unit

(* seq_labels l1 l2 is the label of the edge 
   corresponding to the succession of two edges labeled l1 and l2 *)
  val seq_labels : label -> label -> label
(* merge_labels [ l1 ... ln ] is the set of edges 
   corresponding to the list of parallel edges labeled l1 and l2, 
   after simplification if any is provided *)
  val merge_labels : label list -> label list
end

module type GRAPH = sig
  type t 
  type node
  type label
  type node_description

  module Print : sig
    val f : Format.formatter -> t -> unit
  end

  val description : node -> node_description
  val create_node : node_description -> node
  val eq_node : node -> node -> bool
  val make_empty : unit -> t
  val of_node_descriptions : (node_description list) -> t
  val nodes : t -> node list
  val add_node_of_description : t -> node_description -> unit 
  val add_node : t -> node -> unit 
  val add_edge : t -> node -> label -> node -> unit
  val targets : node -> (node * label) list

  val transitive_close : t -> t
  val order_dag : (node -> node -> int) -> t -> node list
  val restrict_to_degree : t -> label -> unit
end

module Make 
    (Base : LABELED_GRAPH_BASE) : 
    GRAPH with type node_description = Base.node_description 
    and type label = Base.label
	  = struct
	    
(*********************************************************************)
(* Definitions:
   - a graph is a reference on a list of composite nodes
   - a composite node is a record of a node and its targets in the graph
   - a node is a record of a node description (from Base).
   Invariants:
   - nodes are identified physically
   - a node occurs in only one composite node as the source in a graph
   (i.e. no graphs with two composite nodes describing the targets
   of the same node)
   Objectives:
   - the clients must not see the type composite_node
   - provide a function for computing the transitive closure of the graph
 *)
(*********************************************************************)

  type t = node list ref

  and node = 
      { description : node_description ; mutable targets : target list }

  and node_description = Base.node_description

  and label = Base.label 

  and target = { target : node ; mutable label : label }



(******************************************************)
(*                    Debugging                       *)
(******************************************************)

module Print = struct

open Format

let rec f fmt g = 
  fprintf fmt "@[<hv0>%a@]" print_node_list !g ;
  pp_print_newline fmt () ;
  pp_print_flush fmt ()
and print_node_list fmt = 
  List.iter (print_node fmt) 
and print_node fmt { description = desc ; targets = t } =
  fprintf fmt "@[<hv0>%a -> {@[<hv0> %a @]}@] ;@ " 
    Base.print_description desc
    print_targets t
and print_targets fmt = List.iter (print_target fmt) 
and print_target fmt {target = node ; label = l } = 
  fprintf fmt "-%a->%a ; @ " 
    Base.print_label l
    Base.print_description node.description
    
end

(* Basics *)

  let eq_node = (==)
  let description node = node.description

  let of_nodes nodes = ref nodes
  let make_empty () = ref []

  let create_node node_desc = { description = node_desc ; targets = [] }
  let new_node_of_description node_desc = 
    create_node node_desc
  let add_node g node = 
    g := (node :: !g)
  let add_node_of_description g node_desc = 
    add_node g (new_node_of_description node_desc)

  let of_node_descriptions node_descs = 
    of_nodes  (List.map create_node node_descs)
  let nodes g = !g

  let iter = 
    let rec iter_rec f g = 
      match g with
      | [] -> ()
      | node :: g' -> 
	  let _ = (f node) in
	  iter_rec f g'
    in
    fun f g -> (iter_rec f !g)

  let fold = 
    let rec fold_rec f yet g = 
      match g with
      | [] -> yet
      | node :: g' -> 
	  let yet' = (f yet node) in
	  fold_rec f yet' g'
    in
    fun f start g -> (fold_rec f start !g)

  let target_partition node = 
    List.partition 
      (fun { target = node' ; label = label' } -> eq_node node node')

  let add_target targets { target = node ; label = label } =
    let target_node, others = target_partition node targets in
    let labels = Base.merge_labels
	(label :: (List.map (fun target -> target.label) target_node))
    in
    List.fold_left
      (fun targets label' -> { target = node ; label = label' } :: targets)
      others
      labels

  let add_edge g node1 label node2 = 
    node1.targets <- add_target node1.targets { target = node2 ; label = label } ;
    ()

  let copy_dep { target = node ; label = l } = 
    { target = node ; label = l }

  let copy_deps deps = 
    List.map copy_dep deps
      
  let copy_node { description = node_desc ; targets = deps } = 
    { description = node_desc ; targets = copy_deps deps }

  let copy g = 
    ref (List.map (fun node -> copy_node node) !g)

  let copy_nodes_not_edges_and_map g = 
    let new_nodes, assocs =
      List.split  
	(List.map 
	   (fun ({ description = node_desc ; targets = deps } as node) -> 
	     let new_node = { description = node_desc ; targets = [] } in
	     (new_node, (node, new_node))) 
	   !g)
    in
    (ref new_nodes), (fun node -> List.assq node assocs)

  let targets node = 
    List.map (fun { target = t ; label = l } -> (t, l)) node.targets

(******************************************************************)
(*                     Transitive closure                         *)
(******************************************************************)

  let merge_targets ts1 ts2 =
    List.fold_left add_target ts1 ts2

  let prolongate_through_node g node node_targets =
    iter
      (fun node' ->
	let targets' = node'.targets in
	let target_node, other_targets =
	  List.partition
            (fun target -> eq_node node target.target) 
            targets'
	in
	List.iter
	  (fun target ->
            List.iter
              (fun { target = node1 ; label = label1 } ->
		add_edge
		  g node' (Base.seq_labels label1 target.label) node1)
              node_targets)
	  target_node)
      g

  let add_node_targets g node targets =
    List.iter
      (fun { target = node' ; label = label } ->
        add_edge g node label node')
      targets

  let map_target map_nodes { target = node ; label = l } =
    { target = map_nodes node ; label = l }

  let transitive_close g =
    let transitive_g, map_nodes = 
      copy_nodes_not_edges_and_map g 
    in      
    iter
      (fun node ->
	let targets = List.map (map_target map_nodes) node.targets in
	let reachable_from_node_in_transitive_g =
	  fold
	    (fun targets node' ->
	      if List.exists
		  (fun { target = node'' ; label = _ } ->
		    eq_node node' node'')
		  targets
	      then
		let targets' = node'.targets in
		merge_targets targets targets'
	      else targets)
	    targets
	    transitive_g
	in
	let _ = prolongate_through_node 
	    transitive_g (map_nodes node) reachable_from_node_in_transitive_g
	in
	let _ = add_node_targets
	    transitive_g (map_nodes node) reachable_from_node_in_transitive_g 
	in
	())
      g ;
    transitive_g

  let list_removeq guy l = 
    List.find_all (fun guy' -> guy != guy') l

  let list_remove_listq guys l = 
    List.find_all (fun guy' -> not (List.memq guy' guys)) l

  let list_min order l = 
    List.hd (List.sort order l)

  let order_dag default g = 
    let remaining_nodes = ref (!g) in
    let chosen_nodes = ref [] in
    let choose nodes = 
      let only_fathers_nodes = 
	List.fold_left
	  (fun yet node ->
	    let targets = node.targets in
	    let target_nodes = List.map (fun t -> t.target) targets in
	    list_remove_listq target_nodes yet)
	  nodes
	  nodes
      in
      list_min default only_fathers_nodes
    in
    while !remaining_nodes <> [] do
      let chosen = choose !remaining_nodes in
      remaining_nodes := list_removeq chosen !remaining_nodes ;
      chosen_nodes := chosen :: !chosen_nodes 
    done ;
    List.rev !chosen_nodes

  let restrict_to_degree g label = 
    iter
      (fun node ->
	let new_targets = 
	  List.find_all
	    (fun { target = t ; label = label' } ->
	      label' = label)
	    node.targets
	in
	node.targets <- new_targets)
    g

end
