(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)

  
module type LABELED_GRAPH_BASE = sig
  type node_description
  type label

  val print_description : Format.formatter -> node_description -> unit
  val print_label : Format.formatter -> label -> unit

(* seq_labels l1 l2 is the label of the edge 
   corresponding to the succession of two edges labeled l1 and l2 *)
  val seq_labels : label -> label -> label
(* merge_labels [ l1 ... ln ] is the set of edges 
   corresponding to the list of parallel edges labeled l1 and l2, 
   after simplification if any is provided *)
  val merge_labels : label list -> label list
end

module type GRAPH = sig
  type t 
  type node
  type label
  type node_description

  module Print : sig
    val f : Format.formatter -> t -> unit
  end

  val description : node -> node_description
  val create_node : node_description -> node
  val eq_node : node -> node -> bool
  val make_empty : unit -> t
  val of_node_descriptions : (node_description list) -> t
  val nodes : t -> node list
  val add_node_of_description : t -> node_description -> unit 
  val add_node : t -> node -> unit 
  val add_edge : t -> node -> label -> node -> unit
  val targets : node -> (node * label) list

(* Functional transitive closure. *)
  val transitive_close : t -> t

(* Orders a graph without cycles. In case of choice, the first argument f
   serves as a default ordering: if f node1 node2 < 0 then choose node1 else 
   choose node2. 
   In case a cyclic graph is passed as an argument, the exception 
   Failure "hd" is thrown. *)
  val order_dag : (node -> node -> int) -> t -> node list

(* Restrict the argument graph to edges labeled with the second argument.
   Destructive. *)
  val restrict_to_degree : t -> label -> unit

end

module Make : functor (Base : LABELED_GRAPH_BASE) ->
  GRAPH with type node_description = Base.node_description 
  and type label = Base.label
