(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)

(* The lexer definition *)

{
open Parser
(* The table of keywords *)

let create_hashtable size init =
  let tbl = Hashtbl.create size in
  List.iter (fun (key, data) -> Hashtbl.add tbl key data) init;
  tbl

let keyword_table =
  create_hashtable 3 [
  "import" IMPORT ;
  "export" EXPORT ; 
  "end" END ;
  "close" CLOSE ; 
  
  "and", AND ;
  "in", IN ;
  "let", LET
]

}

let blank = [' ' '\010' '\013' '\009' '\012']
let lowercase = ['a'-'z' '\223'-'\246' '\248'-'\255' '_']
let uppercase = ['A'-'Z' '\192'-'\214' '\216'-'\222']
let identchar = 
  ['A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246' '\248'-'\255' '\'' '0'-'9']
let symbolchar =
  ['!' '$' '%' '&' '*' '+' '-' '.' '/' ':' '<' '=' '>' '?' '@' '^' '|' '~']
let decimal_literal = ['0'-'9']+
let hex_literal = '0' ['x' 'X'] ['0'-'9' 'A'-'F' 'a'-'f']+
let oct_literal = '0' ['o' 'O'] ['0'-'7']+
let bin_literal = '0' ['b' 'B'] ['0'-'1']+
let float_literal =
  ['0'-'9']+ ('.' ['0'-'9']* )? (['e' 'E'] ['+' '-']? ['0'-'9']+)?

rule token = parse
    blank +
      { token lexbuf }
  | "@" { AT }
  | "\\" { BACKSLASH }
  | "!" { BANG }
  | "|" { BAR }
  | "," { COMMA }
  | lowercase identchar *
      { let s = Lexing.lexeme lexbuf in
          try
            Hashtbl.find keyword_table s
          with Not_found ->
            LIDENT s }
  | uppercase identchar *
      { UIDENT(Lexing.lexeme lexbuf) }       (* No capitalized keywords *)
  | "#"  { SHARP }
  | "$"  { DOLLAR }
  | "("  { LPAREN }
  | ")"  { RPAREN }
  | "->" { MINUSGREATER }
  | "->>" { MINUSGREATERGREATER }
  | "."  { DOT }
  | "<-" { LESSMINUS }
  | "["  { LBRACKET }
  | "]"  { RBRACKET }
  | "{"  { LBRACE }
  | "|"  { BAR }
  | "}"  { RBRACE }
  | "+"  { PLUS }
  | "="  { EQUAL }
  | "?"  { QUESTION }
  | ";;"  { SEMISEMI }
  | ";"  { SEMI }
  | eof { EOF }
