(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)

type name = Name.t

type degree = Strict | Safe

type ident = definition NamedLocations.identity

and env = (definition, Obj.t) NamedLocations.t

and body = env -> Obj.t

and t = definition list

and definition = 
    { label : ident ;
      mutable body : body ;
      mutable dependencies : dependencies ;
      size : size }

and dependencies = dependency list

and dependency = 
    { target : ident ;
      degree : degree }

and size = 
  | Known of int 
  | Unknown

module Dependencies = struct
  let fold_left = List.fold_left
  let cons x y = x :: y
  let empty = []
end

module Degree = struct
  let min d1 d2 = 
    if (d1 = d2) && (d1 = Safe) then Safe else Strict
end

(* Mixture manipulation *)

let empty = []

let names =
  let rec defined_names_rec yet = function
    | [] -> List.rev yet
    | { label = NamedLocations.Late name ; 
	body = _ ; 
	dependencies = _ ;
	size = _ } 
      :: m -> 
        defined_names_rec (name :: yet)
          m
    | def :: m -> defined_names_rec yet m
  in
  fun m -> defined_names_rec [] m
      
let deferred m =
  let defined = names m in
  List.fold_left 
    (fun yet def ->
      List.fold_left
	(fun deferred { target = target ; degree = _ } ->
	  match target with
	  | NamedLocations.Early _ -> deferred
	  | NamedLocations.Late name ->
	      if List.mem name defined
	      then deferred
	      else LSet.add name deferred)
	yet
	def.dependencies)
    LSet.empty
    m

let fold_left = List.fold_left

let cons def m = 
  match def.label with
  | NamedLocations.Early body -> def :: m
  | NamedLocations.Late name ->
      if List.mem name (names m)
      then failwith 
	  ("Broken invariant : trying to insert two definitions with the same name in a mixin. ")
      else def :: m

let rev = List.rev

(* Module MixGraph: only used at close time for constructing 
   a graph representing a mixin without imports to be instantiated: 
   every dependency can be considered static. The structure should 
   be more compact. And abstract, so the algorithm topologically sorting
   the definitions can be a generic algorithm over graphs. *)


(* For debugging purposes *)
module Print = struct
  open Format
  let rec f fmt m = 
    fprintf fmt "@[<hv2>mix@ @[<hv0>%a@]end@]"
      print_definition_list m
  and print_definition_list fmt = function
    | [] -> ()
    | { label = l ;
	body = _ ;
	dependencies = dependencies } :: m ->
	  fprintf fmt
	    "@[<hv2>%a = _ of@ { @[<hv0>%a@]}@]@ %a"
	    print_label l print_dependencies dependencies
	    print_definition_list m
  and print_label fmt = function
    | NamedLocations.Early _ -> fprintf fmt "_"
    | NamedLocations.Late name -> Name.Print.f fmt name
  and print_dependencies fmt = function
    | [] -> ()
    | dependency :: dependencies ->
	fprintf fmt
	  "@[<hv2>%a@] %a"
	  print_dependency dependency
	  print_dependencies dependencies
  and print_dependency fmt 
      { target = target ; degree = degree } = 
    fprintf fmt 
      "%a : %a" 
      print_target target
      print_degree degree
  and print_target fmt = function
    | NamedLocations.Early _ -> fprintf fmt "<local>"
    | NamedLocations.Late name -> Name.Print.f fmt name
  and print_degree fmt = function
    | Safe -> fprintf fmt "1"
    | Strict -> fprintf fmt "0"
end

module Base = struct
  type node_description = definition
  type label = degree

  let print_description fmt def = 
    Print.print_label fmt def.label

  let print_label = Print.print_degree

  let seq_labels = Degree.min 
  let merge_labels l = 
    [ List.fold_left
	(fun yet dg -> Degree.min yet dg)
	Safe
	l ]
end

module MixGraph = Graph.Make(Base)

let node_named_node_of_definition def = 
  let node = MixGraph.create_node def in
  node, def.label

let make_edge assocs { target = target ; degree = dg } = 
  List.assoc target assocs, dg

(* Generate a graph, thanks to the functions in the interface of DepGraph, 
   knowing that there are no deferred components anymore, so all nodes
   are bound to a definition. *)

(* Correct so that it uses a (node NamedLocations.t) *)
let graph_of_t m = 
  let g = MixGraph.make_empty () in
  let assocs = ref NamedLocations.empty in
  let _ = 
    List.iter 
      (fun def -> 
	let node, id = node_named_node_of_definition def in
	MixGraph.add_node g node ;
	assocs := NamedLocations.extend !assocs id node)
      m
  in
  let _ = 
    List.iter
      (fun 
	({ label = l ; body = b ; dependencies = deps ; size = size } 
	   as def) ->
	     let node = NamedLocations.find def.label !assocs in
	     List.iter
	       (fun { target = target ; degree = dg } -> 
		 let target_node = NamedLocations.find target !assocs in
		 MixGraph.add_edge g target_node dg node)
	       deps)
      m
  in
  g

let extend_env (env : env) def obj = 
  match def.label with
  | NamedLocations.Late name -> NamedLocations.extend_exported env name obj
  | NamedLocations.Early body ->  NamedLocations.extend_local env body obj

let find_env def env = 
  match def.label with
  | NamedLocations.Late name -> NamedLocations.find_exported name env 
  | NamedLocations.Early body ->  NamedLocations.find_local body env

let preallocate defs = 
  let env = ref NamedLocations.empty in
  List.iter
    (fun def -> match def.size with
    | Known n -> 
	let obj = Obj.new_block 0 n in
	env := extend_env !env def obj
    | Unknown -> ())
    defs ;
  !env

let update env def result =
  match def.size with
  | Unknown -> extend_env env def result
  | Known n -> 
      let obj = find_env def env in
      Obj.set_tag obj (Obj.tag result) ;
      for i = 0 to n - 1 do
	Obj.set_field obj i (Obj.field result i)
      done ;
      env

let make_record ordered_defs env = 
  let list = List.fold_left
      (fun yet def ->
	match def.label with
	| NamedLocations.Late name -> 
	    (NamedLocations.find_exported name env) :: yet
	| _ -> yet)
      []
      ordered_defs
  in
  let n = List.length list in
  let res = Obj.new_block 0 n in
  let i = ref (n - 1) in
  List.iter
    (fun obj ->
      Obj.set_field res !i obj ;
      i := !i - 1)
    list ;
  res

let rec defined_before m def1 def2 = match m with
| [] -> raise (Invalid_argument "M.defined_before")
| def :: _ when def == def1 && def == def2 -> 0
| def :: _ when def == def1 -> -1
| def :: _ when def == def2 -> 1
| _ :: rest -> defined_before rest def1 def2

let id_env env ident = NamedLocations.find ident env

let new_dummy_def (yet, mapping) def =
  match def.label with
  | NamedLocations.Early def' ->
      let rec res = { def with label = NamedLocations.Early res }
      in 
      (res :: yet, 
       NamedLocations.extend_local mapping def 
	 (NamedLocations.Early res))
  | NamedLocations.Late name ->
      ({ def with label = NamedLocations.Late name } :: yet, mapping)

let update_dep mapping dep =
    { dep with target = 
      try NamedLocations.find 
	  dep.target 
	  mapping
      with _ -> dep.target }   

let update_def mapping def =
  let body = def.body in
  let dependencies = def.dependencies in
  def.body <-
    begin
      fun env -> body (NamedLocations.compose env mapping)
    end ;
  def.dependencies <-
    List.map
      (update_dep mapping)
      dependencies ;
  ()

let copy m = 
  let dummy_defs, mapping = 
    List.fold_left
      new_dummy_def
      ([], NamedLocations.empty)
      m
  in 
  List.iter
    (fun def -> update_def mapping def)
    dummy_defs ;
  List.rev dummy_defs

let close m = 
  let default_ordering = fun node1 node2 ->
    let def1 = MixGraph.description node1 in
    let def2 = MixGraph.description node2 in
    defined_before m def1 def2 
  in
(* Compute the dep graph g for m *)
  let g = graph_of_t m in
(* Compute \edge{\strict}^+ for g --> g'' *)
  let g' = MixGraph.transitive_close g in
  let _ = MixGraph.restrict_to_degree g' Strict in
(* Compute the preallocation order, and simultaneously 
   take the union with g' *)
  let add_pord = 
    let nodes = MixGraph.nodes g in
    List.iter 
      (fun node -> 
	let targets = MixGraph.targets node in
	List.iter
	  (fun (node', dg) -> 
	    if dg = Strict then () else
	    let desc = MixGraph.description node in
	    if desc.size <> Unknown then () else 
	    MixGraph.add_edge g' node dg node')
	  targets) 
      nodes
  in
  let ordered_defs = 
   List.map 
   MixGraph.description
   (MixGraph.order_dag default_ordering g')
  in
  let init_env = preallocate ordered_defs in
  let final_env = List.fold_left
      (fun env ({ label = l ; body = b ; dependencies = deps ; size = size } as def) ->
	let result = b env in
	update env def result)
      init_env
      ordered_defs
  in
  make_record ordered_defs final_env

let add m1 m2 = 
  let m1 = copy m1 in
  let m2 = copy m2 in
  if List.for_all (fun name1 ->
     List.for_all (fun name2 ->
       name1 <> name2) 
      (names m2))
      (names m1)
  then m1 @ m2
  else failwith 
      "trying to compose two mixins defining a common name. "

let freeze_as name frozen 
    { label = l ; body = b ; dependencies = deps ; size = s } =
  let b' env =
    b (NamedLocations.unfreeze name frozen env)
  in
  let deps' = 
    List.map
      (fun { target = target ; degree = dg } ->
	let target' = match target with
	| NamedLocations.Late name' when name' = name -> 
	    NamedLocations.Early frozen
	| _ -> target
	in
	{ target = target' ; degree = dg })
      deps
  in
  { label = l ; body = b' ; dependencies = deps' ; size = s }  

let freeze name m = 
  let def = List.find (fun def -> def.label = NamedLocations.Late name) m in
  List.map
    (fun def' ->
       freeze_as name def def')
    m
