(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)


open Pcaml;;
open Pa_o;;

exception Mailtoddr

module Var = struct
  type t = string
  let compare = compare
end

module VSet = Set.Make(Var)
let list_union = 
  List.fold_left
    VSet.union
    VSet.empty 
let of_list l = 
  List.fold_left
    (fun yet elt -> VSet.add elt yet)
    VSet.empty 
    l

type src_label = Anonymous of string | Named of string
type def_identity = { src_label : src_label ;    (* Source level label *)
		      src_ident : string ;       (* Source level binder *)
		      trg_ident : string } (* Target definition *)
type dependency = { target : src_label ; degree : M.degree }

let rec fv =
  function
    | <:expr< $e$ . $_$ >> -> fv e
    | <:expr< $anti:e$ >> -> fv e
    | <:expr< $x$ $y$ >> -> VSet.union (fv x) (fv y)
    | <:expr< $e$ . ($f$) >> -> VSet.union (fv e) (fv f)
    | <:expr< [| $list:el$ |] >> -> list_union (List.map fv el)
    | <:expr< $e1$ := $e2$ >> -> VSet.union (fv e1) (fv e2)
    | <:expr< $chr:c$ >> -> VSet.empty
    | <:expr< $flo:s$ >> -> VSet.empty
    | <:expr< for $s$ = $e1$ $to:b$ $e2$ do { $list:el$ } >> ->
	VSet.diff (list_union (List.map fv (e1 :: e2 :: el)))
	  (VSet.singleton s)
    | <:expr< fun [ $list:pwel$ ] >> ->
	list_union 
	  (List.map
	     (fun (p, eop, e) -> 
	       VSet.diff 
		 (VSet.union 
		    (fv e) 
		    (match eop with None -> VSet.empty | Some e' -> fv e')) 
		 (fv_pat p))
	     pwel)
    | <:expr< if $e1$ then $e2$ else $e3$ >> ->
	VSet.union (VSet.union (fv e1) (fv e2))
	  (fv e3)
    | <:expr< $int:s$ >> -> VSet.empty
    | <:expr< let $rec:b$ $list:pel$ in $e$ >> ->
	let keep, throw = 
	  List.fold_left
	    (fun (keep, throw) (p, e) ->
	      (VSet.union keep (fv e), 
	       VSet.union throw (fv_pat p)))
	    (fv e, VSet.empty)
	    pel
	in
	VSet.diff keep throw
    | <:expr< $lid:s$ >> -> VSet.singleton s
    | <:expr< match $e$ with [ $list:pwel$ ] >> ->
	VSet.union 
	  (fv e) 
	  (list_union 
	     (List.map
		(fun (p, eop, e) -> 
		  VSet.diff 
		    (VSet.union 
		       (fv e) 
		       (match eop with None -> VSet.empty | Some e' -> fv e')) 
		    (fv_pat p))
		pwel))
    | <:expr< { $list:eel$ } >> ->
	list_union (List.map (fun (_, e) -> fv e) eel)
    | <:expr< do { $list:el$ } >> ->
	list_union (List.map fv el)
    | <:expr< $e1$ .[ $e2$ ] >> ->
	VSet.union (fv e1) (fv e2)
    | <:expr< $str:s$ >> -> VSet.empty
    | <:expr< try $e$ with [ $list:pwel$ ] >> ->
	VSet.union 
	  (fv e) 
	  (list_union 
	     (List.map
		(fun (p, eop, e) -> 
		  VSet.diff 
		    (VSet.union 
		       (fv e) 
		       (match eop with None -> VSet.empty | Some e' -> fv e')) 
		    (fv_pat p))
		pwel))
    | <:expr< ( $list:el$ ) >> ->
	list_union (List.map fv el)
    | <:expr< ( $e$ : $t$ ) >> -> fv e
    | <:expr< $uid:s$ >> -> VSet.empty
    | <:expr< while $e$ do { $list:el$ } >> ->
	(VSet.union 
	   (fv e) 
	   (list_union (List.map fv el)))
    | _ -> raise Mailtoddr

and fv_pat = function
  | <:patt< $p1$ . $p2$ >> -> VSet.empty
  | <:patt< $anti:e$ >> -> raise Mailtoddr
  | <:patt< ( $p1$ as $p2$ ) >> -> 
      VSet.union (fv_pat p1) (fv_pat p2) 
  | <:patt< _ >> -> VSet.empty
  | <:patt< $p1$ $p2$ >> ->
      VSet.union (fv_pat p1) (fv_pat p2)
  | (<:patt< $chr:s$ >> |
    <:patt< $int:s$ >> |
    <:patt< $str:s$ >>) -> VSet.empty
  | <:patt< $lid:i$ >> -> VSet.singleton i
  | <:patt< $p1$ | $p2$ >> -> fv_pat p1
  | <:patt< $p1$ .. $p2$ >> -> VSet.empty
  | <:patt< { $list:ppl$ } >> -> 
      list_union (List.map (fun (_, e) -> fv_pat e) ppl)
  | <:patt< ( $list:pl$ ) >> -> 
      list_union (List.map fv_pat pl)
  | <:patt< ( $p$ : $t$ ) >> -> fv_pat p
  | <:patt< $uid:s$ >> -> VSet.empty
  | _ -> raise Mailtoddr

let rec is_constructed_value =
  function
    | <:expr< [| $list:el$ |] >> -> true
    | <:expr< fun [ $list:pwel$ ] >> -> true
    | <:expr< { $list:eel$ } >> -> true
    | <:expr< let $rec:b$ $list:pel$ in $e$ >> ->
	is_constructed_value e
    | <:expr< match $e$ with [ $list:pwel$ ] >> ->
	List.for_all (fun (_, _, e) -> is_constructed_value e) pwel
    | <:expr< ( $list:el$ ) >> -> true
    | <:expr< ( $e$ : $t$ ) >> -> is_constructed_value e
    | _ -> false

let count = ref 0
let fresh_name = "mixin___"
let fresh _ = 
  let res = fresh_name ^ (string_of_int !count) in
  incr count ;
  res

let make_label loc exp_ident def_ident = function
  | true -> Named exp_ident
  | false -> Anonymous def_ident

let make_target_label loc exp_ident def_ident = function
  | true -> <:expr<NamedLocations.Late
	$int:(string_of_int (Name.of_string exp_ident))$>>
  | false -> <:expr<NamedLocations.Early $lid:def_ident$>>

let make_dependency loc lab id defid dg = 
  { target = lab ;
    degree = dg }

let dependencies loc fvs ii_vars def_descriptions expr fake_deps = 
  let dg =  
    if is_constructed_value expr 
    then M.Safe
    else M.Strict
  in
  let def_desc_dep_of_id id = 
    try
      let 
	  { src_label = l ; src_ident = id' ; trg_ident = defid } 
	  as def_desc =
	(List.find
	   (fun def -> def.src_ident = id)
	   def_descriptions)
      in
      (def_desc, make_dependency loc l id defid dg) 
    with Not_found -> 
      ({ src_label = Named id ; 
	 src_ident = id ; 
	 trg_ident = id },
       { target = Named id ; degree = dg }) 
  in
  let fvset = 
    VSet.inter
      fvs
      ii_vars
  in
  let fake_deps = 
    List.map
      (fun id -> def_desc_dep_of_id id)
      fake_deps
  in
  VSet.fold
    (fun id yet -> (def_desc_dep_of_id id) :: yet)
    fvset
    fake_deps

let convert_degree loc = function
  | M.Safe -> <:expr<M.Safe>>
  | M.Strict -> <:expr<M.Strict>>

let convert_identity loc identity = 
  match identity with
  | Named s ->
      <:expr<NamedLocations.Late
	$int:(string_of_int (Name.of_string s))$>>
  | Anonymous s -> 
      <:expr<NamedLocations.Early $lid:s$>>

let convert_dep loc { target = identity ; degree = dg } = 
  <:expr<{ M.target = $convert_identity loc identity$ ;
	   M.degree = $convert_degree loc dg$ }>>

let rec lookup_env context env_name deps expr = 
  let loc = MLast.loc_of_expr expr in
  match deps with 
  | [] -> context (<:expr<(Obj.repr ($expr$))>>)
  | (def_desc, dep) :: deps ->
      let id = def_desc.src_ident in
      let access = 
	match dep.target with
	| Named name ->
	    <:expr<Obj.magic 
	      (NamedLocations.find_exported 
		 $int:(string_of_int (Name.of_string name))$ $lid:env_name$)>>
	| Anonymous def_id ->
	    <:expr<Obj.magic
	      (NamedLocations.find_local $lid:def_id$ 
					       $lid:env_name$)>>
      in
      let context' = fun x ->
	<:expr<let $lid:id$ = $access$ in
        $context x$>>
      in lookup_env context' env_name deps expr
	
let size_of_expr loc fvs = function
  | <:expr< fun [ $list:pwel$ ] >> -> 
      <:expr<M.Known $int:(string_of_int (1 + (VSet.cardinal fvs)))$>>
  | <:expr< { $list:eel$ } >> -> 
      <:expr<M.Known $int:(string_of_int (VSet.cardinal fvs))$>>
  | _ -> <:expr<M.Unknown>>

let id_of_def_desc loc  = 
  function def_desc -> <:expr<$lid:def_desc.trg_ident$>>


let rec make_list loc = function
  | [] -> <:expr<[]>>
  | a :: l -> <:expr<[$a$ :: $(make_list loc l)$]>>

let definition = Grammar.Entry.create gram "definition";;
let fake_dependencies = Grammar.Entry.create gram "fake_dependencies";;
let binding = Grammar.Entry.create gram "binding";;

EXTEND
  binding:
    [ RIGHTA
      [ p = patt LEVEL "simple"; e = SELF -> <:expr< fun $p$ -> $e$ >>
      | "="; e = expr -> <:expr< $e$ >>
      | ":"; t = ctyp; "="; e = expr -> <:expr< ($e$ : $t$) >> ] ]
  ;

  fake_dependencies:
  [ [ "of" ; "{" ; deps = LIST0 LIDENT ; "}" -> deps ] ] ;

  definition: 
  [ [  "let" ; ident = LIDENT ; fake_deps = OPT fake_dependencies ; 
       e = binding -> 
	 (true, ident, e, 
	  (match fake_deps with None -> [] | Some x -> x)) 
    |  "local" ; ident = LIDENT ; fake_deps = OPT fake_dependencies ; 
        e = binding -> 
	  (false, ident, e, 
	  (match fake_deps with None -> [] | Some x -> x)) 
  ] ] ;

  expr: AFTER "apply"
  [ [ "import" ; ii = LIST0 LIDENT ; 
      "export" ; out = LIST0 definition ; "end" ->
	begin
          let n = List.length out in
	  let def_descriptions = List.map 
	      (fun (lab, ident, expr, fake_deps) ->
		let ident_def = fresh () in
		let loc = MLast.loc_of_expr expr in
		{ src_label = make_label loc ident ident_def lab ;
		  src_ident = ident ; 
		  trg_ident = ident_def })
	      out
	  in
	  let ii_vars = 
	    VSet.union
	      (of_list ii)
	      (of_list 
		 (List.map (fun def -> def.src_ident) def_descriptions))
	  in
	  let defs = List.map2
	      (fun (local, ident, expr, fake_deps) def_descr ->
		let fvs = fv expr in
		let deps = 
		  dependencies loc fvs ii_vars def_descriptions expr fake_deps
		in
		let the_label = 
		  make_target_label loc ident def_descr.trg_ident local
		in
		let (e : MLast.expr) = 
		  <:expr< { M.label = 
			    $the_label$ ;
			    M.body = 
			    fun [ $lid:"env__0"$ -> 
			      $(lookup_env
				  (fun x -> x) 
				  "env__0" 
				  deps 
				  expr)$ ] ;
			      M.size = $(size_of_expr loc fvs expr)$ ;
			      M.dependencies = $make_list loc
				(List.map (convert_dep loc)
				   (snd (List.split deps)))$ 
			  } >>
		in
		let p = <:patt<$lid:(def_descr.trg_ident)$>>in
		(p, e))
	      out
	      def_descriptions
	  in
	  <:expr<let rec $list:defs$ in 
	  $make_list loc (List.map (id_of_def_desc loc) def_descriptions)$>> 
	end
  | "close" ; e = expr ; "at" ; t = ctyp -> 
      <:expr<(Obj.magic (M.close $e$) : $t$)>>
  | e1 = expr ; "<+>" ; e2 = expr ->
      <:expr<M.add $e1$ $e2$>>
  ] ] ;

END
