(***********************************************************************)
(*                                                                     *)
(*                                CMSV                                 *)
(*                                                                     *)
(*                  Tom Hirschowitz, INRIA Rocquencourt                *)
(*                                                                     *)
(*  Copyright 2003 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU General Public License.                 *)
(*                                                                     *)
(***********************************************************************)


type expr = 
  | Var of string
  | Lam of string * expr
  | App of expr * expr
  | Let of (bindings * expr)
  | Num of int
  | String of string
  | Op of operator * (expr list)
and valexpr = 
  | Int of int
  | Str of string
  | Clo of string * expr
and bindings = (string * expr) list
and env = (string * valexpr) list
and operator = 
  | Print | Plus 

exception Runtime_error

let eval = 
  import 
    bind access eval app_operator
  export
    let eval env = function
      | Var s -> access s env
      | Lam (s, e) -> Clo (s, e)
      | App(e1, e2) ->
	  let v1 = eval env e1 in
	  begin
	    match v1 with
	    | Clo (s, e) -> 
		eval (bind [(s, e2)] env) e
	    | _ -> raise Runtime_error
	  end
      | Let (b, e) -> eval (bind b env) e
      | Num n -> Int n
      | String s -> Str s 
      | Op (o, args) ->
	  let vargs = List.map (eval env) args in
	  app_operator env o vargs
end

let env = 
  import 
    eval
  export 
    let access s env = 
      try List.find s env with Not_found -> raise Runtime_error
    let extend s v env = (s, v) :: env
    let empty = []
end

let bindings = 
  import 
    extend eval
  export
    let bind l env =
      List.fold_left
	(fun env (s, e) ->
	  let v = eval env e in
	  extend s v env)
	env
	l
end

let operators = 
  import
  export
    let app_operator env o args = match o, args with
    | Print, [v] ->
	begin
	  match v with
	  | Clo (_,_) -> print_string " <fun> "
	  | Int n -> print_int n 
	  | Str s -> print_string s
	end ;
	Int 0
    | Plus, [ Int n1 ; Int n2] -> 
	Int (n1 + 2)  
    | _, _ -> raise Runtime_error
end

let test = 
  import 
    empty eval
  export
    local test = 
      eval empty
        (Let 
	   ([ ("x", Op(Print, [String "un entier :"])) ;
	      ("y", Op(Print, [
		       Op(Plus, [Num 1 ; Num 2]) ])) ;
	      ("_", Op(Print, [String "\n"])) ],
	    Num 0))
end

let open_evaluator = eval <+> env <+> bindings <+> operators <+> test

type t = 
    { eval : env -> expr -> valexpr ;
      access : string -> env -> valexpr ;
      extend : string -> valexpr -> env -> env ;
      empty : env ; 
      bind : bindings -> env -> env ;
      app_operator : env -> operator -> (valexpr list) -> valexpr }

let evaluator = close open_evaluator at t

